package it.inaf.oats.ia2.tapschemamanager.datalayer;

import java.io.Serializable;
import java.math.BigInteger;
import javax.persistence.Column;
import javax.persistence.Entity;
import javax.persistence.FetchType;
import javax.persistence.Id;
import javax.persistence.IdClass;
import javax.persistence.ManyToOne;
import javax.persistence.PrimaryKeyJoinColumn;
import javax.persistence.Table;

class ColumnId implements Serializable {

    private static final long serialVersionUID = 6943729649657196881L;

    String fullTableName;
    String name;
}

/**
 * JPA entity for the table <code>TAP_SCHEMA.columns</code>.
 *
 * @author Sonia Zorba {@literal <zorba at oats.inaf.it>}
 */
@Entity
@IdClass(ColumnId.class)
@Table(name = "columns")
public class ColumnEntity implements Serializable, TapSchemaEntity {

    private static final long serialVersionUID = 3630294743516824359L;

    @Id
    @Column(name = "table_name", length = 128)
    private String fullTableName;

    @Id
    @Column(name = "column_name", length = 64)
    private String name;

    @ManyToOne(fetch = FetchType.EAGER)
    @PrimaryKeyJoinColumn(name = "table_name", referencedColumnName = "table_name")
    private TableEntity table;

    @Column(name = "utype", length = 512)
    private String utype;

    @Column(name = "ucd", length = 64)
    private String ucd;

    @Column(name = "unit", length = 64)
    private String unit;

    @Column(name = "description", length = 512)
    private String description;

    @Column(name = "datatype", length = 64)
    private String datatype;

    @Column(name = "size")
    private int size;

    @Column(name = "principal")
    private int principal;

    @Column(name = "indexed")
    private int indexed;

    @Column(name = "std")
    private int std;

    @Column(name = "columnID")
    private BigInteger columnID;

    @Column(name = "id")
    private int id;

    private boolean primaryKey;

    protected ColumnEntity() {
    }

    protected ColumnEntity(String columnName) {
        this();
        this.name = columnName;
    }

    /**
     * The complete table name:
     * <code>&lt;schema-name&gt;.&lt;table-name&gt;</code>.
     */
    public String getFullTableName() {
        return fullTableName;
    }

    protected void setFullTableName(String fullTableName) {
        this.fullTableName = fullTableName;
    }

    /**
     * The value in the <code>column_name</code> column.
     */
    @Override
    public String getName() {
        return name;
    }

    protected void setName(String columnName) {
        this.name = columnName;
    }

    /**
     * The value in the <code>utype</code> column.
     */
    public String getUtype() {
        return utype;
    }

    public void setUtype(String utype) {
        this.utype = utype;
    }

    /**
     * The value in the <code>ucd</code> column: represents the Unified Content
     * Descriptor of the column.
     */
    public String getUcd() {
        return ucd;
    }

    public void setUcd(String ucd) {
        this.ucd = ucd;
    }

    /**
     * The value in the <code>unit</code> column.
     */
    public String getUnit() {
        return unit;
    }

    public void setUnit(String unit) {
        this.unit = unit;
    }

    /**
     * The value in the <code>description</code> column.
     */
    public String getDescription() {
        return description;
    }

    public void setDescription(String description) {
        this.description = description;
    }

    /**
     * The value in the <code>datatype</code> column.
     */
    public String getDatatype() {
        return datatype;
    }

    public void setDatatype(String datatype) {
        this.datatype = datatype;
    }

    /**
     * The value in the <code>size</code> column.
     */
    public int getSize() {
        return size;
    }

    public void setSize(int size) {
        this.size = size;
    }

    /**
     * The value in the <code>principal</code> column: indicates that the column
     * is considered a core part the content.<br> It is an integer but should be
     * treated as a boolean value.
     */
    public int getPrincipal() {
        return principal;
    }

    public void setPrincipal(int principal) {
        this.principal = principal;
    }

    /**
     * The value in the <code>indexed</code> column: indicates that the column
     * is indexed.<br> It is an integer but should be treated as a boolean
     * value.
     */
    public int getIndexed() {
        return indexed;
    }

    public void setIndexed(int indexed) {
        this.indexed = indexed;
    }

    /**
     * The value in the <code>std</code> column: indicates that the column is
     * defined by some standard.<br> It is an integer but should be treated as a
     * boolean value.
     */
    public int getStd() {
        return std;
    }

    public void setStd(int std) {
        this.std = std;
    }

    /**
     * The value in the <code>columnID</code> column: it represents [TODO]...
     */
    protected BigInteger getColumnID() {
        return columnID;
    }

    protected void setColumnID(BigInteger columnID) {
        this.columnID = columnID;
    }

    /**
     * The {@link TableEntity} that owns this ColumnEntity.
     */
    public TableEntity getTable() {
        return table;
    }

    protected void setTable(TableEntity tableEntity) {
        this.table = tableEntity;
    }

    public int getId() {
        return id;
    }

    public void setId(int id) {
        this.id = id;
    }

    /**
     * Indicates if the column is a primary key (or a part of a primary key).
     * This information is not stored into the TAP_SCHEMA, so it will be
     * reloaded from the source schema each time.
     */
    public boolean isPrimaryKey() {
        return primaryKey;
    }

    public void setPrimaryKey(boolean primaryKey) {
        this.primaryKey = primaryKey;
    }
}
