package it.inaf.oats.ia2.tapschemamanager.datalayer;

import java.io.Serializable;
import java.math.BigInteger;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import javax.persistence.CascadeType;
import javax.persistence.Column;
import javax.persistence.Entity;
import javax.persistence.Id;
import javax.persistence.MapKeyColumn;
import javax.persistence.OneToMany;
import javax.persistence.Table;

/**
 * JPA entity for the table <code>TAP_SCHEMA.schemas</code>.
 *
 * @author Sonia Zorba {@literal <zorba at oats.inaf.it>}
 */
@Entity
@Table(name = "\"schemas\"")
public class SchemaEntity implements Serializable, TapSchemaEntity {

    private static final long serialVersionUID = -1058919664128696098L;

    @Id
    @Column(name = "schema_name", length = 64)
    private String name;

    @Column(name = "utype", length = 512)
    private String utype;

    @Column(name = "description", length = 512)
    private String description;

    @Column(name = "schemaID")
    private BigInteger schemaID;

    @OneToMany(mappedBy = "schema", cascade = CascadeType.ALL, orphanRemoval = true)
    @MapKeyColumn(name = "table_name", insertable = false, updatable = false)
    private final List<TableEntity> tables;

    protected SchemaEntity() {
        this.tables = new ArrayList<>();
    }

    protected SchemaEntity(String schemaName) {
        this();
        this.name = schemaName;
    }

    /**
     * The value in the <code>schema_name</code> column.
     */
    @Override
    public String getName() {
        return name;
    }

    protected void setName(String schemaName) {
        this.name = schemaName;
    }

    /**
     * The value in the <code>utype</code> column.
     */
    public String getUtype() {
        return utype;
    }

    public void setUtype(String utype) {
        this.utype = utype;
    }

    /**
     * The value in the <code>description</code> column.
     */
    public String getDescription() {
        return description;
    }

    public void setDescription(String description) {
        this.description = description;
    }

    /**
     * The value in the <code>schemaID</code> column: it represents [TODO]...
     */
    protected BigInteger getSchemaID() {
        return schemaID;
    }

    protected void setSchemaID(BigInteger schemaID) {
        this.schemaID = schemaID;
    }

    /**
     * The list of <code>TableEntity</code> owned by this
     * <code>SchemaEntity</code>.
     *
     * @return an unmodifiable list
     */
    public List<TableEntity> getTables() {
        return Collections.unmodifiableList(tables);
    }

    /**
     * Find the <code>TableEntity</code> owned by this <code>SchemaEntity</code>
     * using its short name (the table name without the schema name).
     *
     * @return the <code>TableEntity</code> instance if it exists, null
     * otherwise.
     */
    public TableEntity getTableByShortName(String shortTableName) {
        return getTable(name + "." + shortTableName);
    }

    /**
     * Find the <code>TableEntity</code> owned by this <code>SchemaEntity</code>
     * using its complete name (<code>schema_name.table_name</code>).
     *
     * @return the <code>TableEntity</code> instance if it exists, null
     * otherwise.
     */
    public TableEntity getTable(String fullTableName) {
        return DLUtil.getEntityByName(tables, fullTableName);
    }

    protected void addTable(TableEntity table) {
        tables.add(table);
        table.setSchemaName(name);
        table.setSchema(this);
    }

    protected void removeTable(String fullTableName) {
        DLUtil.removeEntity(tables, fullTableName);
    }
}
