package it.inaf.oats.ia2.tapschemamanager.datalayer.test;

import it.inaf.oats.ia2.tapschemamanager.datalayer.ColumnEntity;
import it.inaf.oats.ia2.tapschemamanager.datalayer.Credentials;
import it.inaf.oats.ia2.tapschemamanager.datalayer.DBWrapper;
import it.inaf.oats.ia2.tapschemamanager.datalayer.DLUtil;
import it.inaf.oats.ia2.tapschemamanager.datalayer.DatabaseType;
import it.inaf.oats.ia2.tapschemamanager.datalayer.KeyColumnEntity;
import it.inaf.oats.ia2.tapschemamanager.datalayer.KeyEntity;
import it.inaf.oats.ia2.tapschemamanager.datalayer.SchemaEntity;
import it.inaf.oats.ia2.tapschemamanager.datalayer.TableEntity;
import it.inaf.oats.ia2.tapschemamanager.datalayer.TapSchemaHandler;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.List;
import org.junit.After;
import org.junit.AfterClass;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Test;
import static org.junit.Assert.*;
import org.junit.Ignore;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 *
 * @author Sonia Zorba {@literal <zorba at oats.inaf.it>}
 */
public class TestQuery {

    private static final Logger log = LoggerFactory.getLogger(TestQuery.class);

    private static final int SCHEMAS_COUNT = 2; // minimum value: 2
    private static final int TABLES_COUNT = 3;

    private static List<DBWrapper> dbWrappers;

    public TestQuery() {
    }

    @BeforeClass
    public static void setUpClass() throws SQLException {

        dbWrappers = new ArrayList<>();

        // MYSQL
        Credentials mysqlCredentials = new Credentials(DatabaseType.MYSQL);
        mysqlCredentials.setHostname("localhost");
        mysqlCredentials.setUsername("root");
        mysqlCredentials.setPassword("root");

        // POSTGRES
        Credentials postgresCredentials = new Credentials(DatabaseType.POSTGRES);
        postgresCredentials.setHostname("localhost");
        postgresCredentials.setUsername("postgres");
        postgresCredentials.setPassword("pippo");

        DBWrapper dbWrapper = new DBWrapper(mysqlCredentials);
        dbWrapper.testConnections();
        //dbWrappers.add(dbWrapper);

        dbWrapper = new DBWrapper(postgresCredentials);
        dbWrapper.testConnections();
        dbWrappers.add(dbWrapper);

        // Mix!
        //dbWrappers.add(new DBWrapper(mysqlCredentials, postgresCredentials));
        //dbWrappers.add(new DBWrapper(postgresCredentials, mysqlCredentials));
    }

    @AfterClass
    public static void tearDownClass() {
    }

    @Before
    public void setUp() {
    }

    @After
    public void tearDown() {
    }

    private void removeTestingDatabases() throws SQLException {
        for (DBWrapper dbWrapper : dbWrappers) {
            try (Connection sourceConnection = dbWrapper.getSourceConnection();
                    Statement statement = sourceConnection.createStatement()) {
                DatabaseType dbType = dbWrapper.getSourceDatabaseType();

                // Removing keys between schema1 and schema0
                if (dbType == DatabaseType.MYSQL) {
                    try (ResultSet rs = statement.executeQuery("SHOW DATABASES WHERE `Database` = 'sch1'")) {
                        // The constraint can be removed only if sch1 exists
                        if (rs.next()) {
                            statement.executeUpdate("ALTER TABLE sch1.table0 DROP FOREIGN KEY sch0table0id_constraint");
                        }
                    }
                } else if (dbType == DatabaseType.POSTGRES) {
                    try (ResultSet rs = statement.executeQuery("SELECT schema_name FROM information_schema.schemata WHERE schema_name = 'sch1'")) {
                        // The constraint can be removed only if sch1 exists
                        if (rs.next()) {
                            statement.executeUpdate("ALTER TABLE sch1.table0 DROP CONSTRAINT sch0table0id_constraint");
                        }
                    }
                } else {
                    throw new UnsupportedOperationException("Database type " + dbType + " not supported");
                }

                for (int i = 0; i < SCHEMAS_COUNT; i++) {
                    if (dbType == DatabaseType.MYSQL) {
                        statement.executeUpdate("DROP DATABASE IF EXISTS sch" + i);
                    } else if (dbType == DatabaseType.POSTGRES) {
                        statement.executeUpdate("DROP SCHEMA IF EXISTS sch" + i + " CASCADE");
                    } else {
                        throw new UnsupportedOperationException("Database type " + dbType + " not supported");
                    }
                }
            }
            try (Connection tapSchemaConnection = dbWrapper.getTapSchemaConnection();
                    Statement statement = tapSchemaConnection.createStatement()) {
                DatabaseType dbType = dbWrapper.getTapSchemaDatabaseType();
                if (dbType == DatabaseType.MYSQL) {
                    statement.executeUpdate("DROP DATABASE IF EXISTS test_tap_schema");
                } else if (dbType == DatabaseType.POSTGRES) {
                    statement.executeUpdate("DROP SCHEMA IF EXISTS test_tap_schema CASCADE");
                } else {
                    throw new UnsupportedOperationException("Database type " + dbType + " not supported");
                }
            }
        }
    }

    private void setUpTestingDatabases() throws SQLException {
        removeTestingDatabases();

        for (DBWrapper dbWrapper : dbWrappers) {

            DatabaseType dbType = dbWrapper.getSourceDatabaseType();

            try (Connection connection = dbWrapper.getSourceConnection()) {

                try (Statement statement = connection.createStatement()) {

                    if (dbType == DatabaseType.MYSQL) {
                        for (int i = 0; i < SCHEMAS_COUNT; i++) {
                            statement.executeUpdate("CREATE DATABASE sch" + i);
                            for (int j = 0; j < TABLES_COUNT; j++) {
                                statement.executeUpdate("CREATE TABLE sch" + i + ".table" + j + " (\n"
                                        + "id INT PRIMARY KEY AUTO_INCREMENT,\n"
                                        + "value1 VARCHAR(255),\n"
                                        + "value2 FLOAT\n"
                                        + ");");
                            }
                        }
                    } else if (dbType == DatabaseType.POSTGRES) {
                        for (int i = 0; i < SCHEMAS_COUNT; i++) {
                            statement.executeUpdate("CREATE SCHEMA sch" + i);
                            for (int j = 0; j < TABLES_COUNT; j++) {
                                statement.executeUpdate("CREATE TABLE sch" + i + ".table" + j + " (\n"
                                        + "id BIGSERIAL PRIMARY KEY,\n"
                                        + "value1 VARCHAR(255),\n"
                                        + "value2 FLOAT\n"
                                        + ");");
                            }
                        }
                    } else {
                        throw new UnsupportedOperationException("Database type " + dbType + " not supported");
                    }
                }

                log.info("dbs created");

                try (Statement statement = connection.createStatement()) {
                    for (int j = 0; j < TABLES_COUNT - 1; j++) {
                        String update1 = "ALTER TABLE sch0.table" + (j + 1) + " ADD COLUMN table" + j + "_id INT";
                        statement.executeUpdate(update1);
                        String update2;
                        if (dbType == DatabaseType.MYSQL || dbType == DatabaseType.POSTGRES) {
                            update2 = "ALTER TABLE sch0.table" + (j + 1) + " ADD CONSTRAINT id_constraint_" + (j + 1) + " FOREIGN KEY (table" + j + "_id) REFERENCES sch0.table" + j + "(id)";
                        } else {
                            throw new UnsupportedOperationException("Database type " + dbType + " not supported");
                        }

                        statement.executeUpdate(update2);
                    }

                    // Foreign key with multiple columns
                    statement.executeUpdate("CREATE TABLE sch0.tableX (idx1 INT, idx2 INT)");
                    statement.executeUpdate("ALTER TABLE sch0.tableX ADD CONSTRAINT pkx PRIMARY KEY (idx1, idx2)");
                    statement.executeUpdate("CREATE TABLE sch0.tableY (idy1 INT, idy2 INT)");
                    statement.executeUpdate("ALTER TABLE sch0.tableY ADD CONSTRAINT pky PRIMARY KEY (idy1, idy2)");
                    statement.executeUpdate("ALTER TABLE sch0.tableY ADD CONSTRAINT fky FOREIGN KEY(idy1, idy2) REFERENCES sch0.tableX(idx1, idx2)");

                    // Foreign keys between different schemas
                    statement.executeUpdate("ALTER TABLE sch0.table0 ADD COLUMN sch1table0id INT");
                    statement.executeUpdate("ALTER TABLE sch0.table0 ADD CONSTRAINT sch1table0id_constraint FOREIGN KEY(sch1table0id) REFERENCES sch1.table0(id)");
                    statement.executeUpdate("ALTER TABLE sch1.table0 ADD COLUMN sch0table0id INT");
                    statement.executeUpdate("ALTER TABLE sch1.table0 ADD CONSTRAINT sch0table0id_constraint FOREIGN KEY(sch0table0id) REFERENCES sch0.table0(id)");
                }
            }
        }
    }

    @Test
    public void testTapSchemaSerialization() throws Exception {
        for (DBWrapper dbWrapper : dbWrappers) {
            if (dbWrapper.getTapSchemaDatabaseType() == DatabaseType.MYSQL) { // currently "tng_TAP_SCHEMA" exists in a MySQL instance
                TapSchemaHandler tapSchemaHandler = new TapSchemaHandler(dbWrapper, "tng_TAP_SCHEMA", true);

                log.info("schemas size = {}", tapSchemaHandler.getSchemas().size());

                try (FileOutputStream fileOut = new FileOutputStream("/home/sonia/test.ser");
                        ObjectOutputStream out = new ObjectOutputStream(fileOut)) {
                    out.writeObject(tapSchemaHandler);
                }

                try (FileInputStream fileIn = new FileInputStream("/home/sonia/test.ser");
                        ObjectInputStream in = new ObjectInputStream(fileIn)) {
                    tapSchemaHandler = (TapSchemaHandler) in.readObject();
                }

                tapSchemaHandler.save();
                log.info("schemas size = {}", tapSchemaHandler.getSchemas().size());
            }
        }
    }

    //@Ignore
    @Test
    public void loadSchema() throws SQLException {
        for (DBWrapper dbWrapper : dbWrappers) {
            if (dbWrapper.getTapSchemaDatabaseType() == DatabaseType.MYSQL) { // currently "tng_TAP_SCHEMA" exists in a MySQL instance
                TapSchemaHandler tapSchemaHandler = new TapSchemaHandler(dbWrapper, "tng_TAP_SCHEMA", true);
                assertTrue(!tapSchemaHandler.getSchemas().isEmpty());

                // Testing alphabetical order loading
                for (SchemaEntity schemaEntity : tapSchemaHandler.getSchemas()) {
                    TableEntity previousTable = null;
                    for (TableEntity table : schemaEntity.getTables()) {
                        if (previousTable != null) {
                            //assertTrue("Table=" + table.getName() + ", Previous Table=" + previousTable.getName(),
                            //        previousTable.getName().compareToIgnoreCase(table.getName()) < 1);
                        }
                        previousTable = table;

                        ColumnEntity previousColumn = null;
                        for (ColumnEntity column : table.getColumns()) {
                            if (previousColumn != null) {
                                //assertTrue("Column=" + column.getName() + ", Previous Column=" + previousColumn.getName(),
                                //        previousColumn.getName().compareToIgnoreCase(column.getName()) < 1);
                            }
                            previousColumn = column;
                        }

//                if(previousTable == null) {
//                    previousTable = table;
//                } else {
//                }
                    }
                }
            }
        }

    }

    private void printKeys(String message, List<KeyEntity> keys) {
        log.info("{}: Found {} keys", message, keys.size());
        int index = 0;
        for (KeyEntity key : keys) {
            for (KeyColumnEntity keyColumn : key.getKeyColumns()) {
                log.info("[{}] {}.{} -> {}.{}", index, key.getFromTableFullName(), keyColumn.getFromColumn(), key.getTargetTableFullName(), keyColumn.getTargetColumn());
            }
            index++;
        }
    }

    //@Ignore
    @Test
    public void createNewAndUpdate() throws SQLException {
        log.info("TEST createNewAndUpdate STARTED");

        try {
            removeTestingDatabases();

            setUpTestingDatabases();

            for (DBWrapper dbWrapper : dbWrappers) {

                // Initializing a not existing TAP_SCHEMA
                TapSchemaHandler tsh = new TapSchemaHandler(dbWrapper, "test_tap_schema", false);

                /////////////////////////////////////
                //         ADDING A SCHEMA         //
                /////////////////////////////////////
                //
                SchemaEntity sch0 = tsh.getNewSchema("sch0");
                tsh.addSchema(sch0);

                List<KeyEntity> sch0Keys = tsh.getAllSchemaKeys("sch0");
                printKeys("ALL sch0 keys", sch0Keys);

                // In the testing schemas each numbered table references the id 
                // of the previous table, except for the first table, so there
                // should be "TABLES_COUNT - 1" keys for the numbered tables.
                // Moreover, there are also the following keys:
                // - sch0.table0.sch1table0id -> sch1.table0.id
                // - sch1.table0.sch0table0id -> sch0.table0.id
                // - sch0.tableY.(idy1, idy2) -> sch0.tableX.(idyx, idx2)
                // so we check for TABLES_COUNT + 2.
                assertEquals(sch0Keys.size(), TABLES_COUNT + 2);

                // Checking that keys information has been filled.
                for (KeyEntity schemaKey : sch0Keys) {
                    assertNotNull(schemaKey.getFromTableFullName());
                    assertNotNull(schemaKey.getTargetTableFullName());
                    for (KeyColumnEntity keyColumn : schemaKey.getKeyColumns()) {
                        assertNotNull(keyColumn.getFromColumn());
                        assertNotNull(keyColumn.getTargetColumn());
                    }
                }

                /////////////////////////////////////
                //         ADDING A TABLE          //
                /////////////////////////////////////
                //
                TableEntity sch0table0 = tsh.getNewTable("sch0", "table0");
                assertEquals(sch0table0.getName(), "sch0.table0");
                tsh.addTable("sch0", sch0table0);
                assertEquals(sch0.getTables().size(), 1);

                // Checking that columns information has been filled.
                List<ColumnEntity> sch0table0Columns = tsh.getColumnEntities("sch0", "table0");
                for (ColumnEntity ce : sch0table0Columns) {
                    assertNotNull(ce.getTable());
                }
                // Primary key check
                assertTrue(DLUtil.getEntityByName(sch0table0Columns, "id").isPrimaryKey());

                /////////////////////////////////////
                //         KEYS MANAGEMENT         //
                /////////////////////////////////////
                //
                // CASE 1: Foreign key between two tables in the same schema.
                // sch0.table1.t0id -> sch0.table0.id
                //
                // Adding sch0.table1
                TableEntity sch0table1 = tsh.getNewTable("sch0", "table1");
                tsh.addTable("sch0", sch0table1);
                assertEquals(sch0.getTables().size(), 2);

                printKeys("sch0table0 fromKeys", sch0table0.getFromKeys());
                assertTrue(sch0table0.getFromKeys().isEmpty());
                assertEquals(sch0table0.getTargetKeys().size(), 1);
                assertEquals(sch0table0.getTargetKeys().get(0).getFromTableFullName(), "sch0.table1");
                assertEquals(sch0table0.getTargetKeys().get(0).getTargetTableFullName(), "sch0.table0");
                assertEquals(sch0table1.getFromKeys().size(), 1);
                assertEquals(sch0table1.getFromKeys().get(0).getFromTableFullName(), "sch0.table1");
                assertEquals(sch0table1.getFromKeys().get(0).getTargetTableFullName(), "sch0.table0");
                assertTrue(sch0table1.getTargetKeys().isEmpty());

                // Removing sch0.table1
                tsh.removeTable(sch0, sch0table1);
                assertEquals(sch0.getTables().size(), 1);
                assertTrue(sch0table0.getFromKeys().isEmpty());
                assertTrue(sch0table0.getTargetKeys().isEmpty());

                //
                // CASE 2: Foreign key between two tables in different schemas.
                // sch1.table0.sch0table0id -> sch0.table0.id
                //
                // Adding sch1
                SchemaEntity sch1 = tsh.getNewSchema("sch1");
                tsh.addSchema(sch1);
                List<KeyEntity> sch1Keys = tsh.getAllSchemaKeys("sch1");
                assertEquals(sch1Keys.size(), 2);
                for (KeyEntity keyEntity : sch1Keys) {
                    String fromSchemaName = keyEntity.getFromSchemaName();
                    String targetSchemaName = keyEntity.getTargetSchemaName();
                    assertTrue(fromSchemaName.equals(""));
                    // TODO <---
                }

                // Adding sch1.table0
                TableEntity sch1table0 = tsh.getNewTable("sch1", "table0");
                tsh.addTable("sch1", sch1table0);

                assertEquals(sch0table0.getTargetKeys().size(), 1);
                assertEquals(sch0table0.getTargetKeys().get(0).getFromTableFullName(), "sch1.table0");
                assertEquals(sch0table0.getTargetKeys().get(0).getTargetTableFullName(), "sch0.table0");
                assertEquals(sch1table0.getFromKeys().size(), 1);
                assertEquals(sch0table0.getFromKeys().get(0).getFromTableFullName(), "sch1.table0");
                assertEquals(sch0table0.getFromKeys().get(0).getTargetTableFullName(), "sch0.table0");

                // Removing sch1
                tsh.removeSchema("sch1");
                assertTrue(sch0table0.getTargetKeys().isEmpty());
                assertTrue(sch1table0.getFromKeys().isEmpty());

                // Case 2B: Re-adding the same instance of sch1
                // sch1.table0 has not been removed from its schema, so the keys
                // should be re-added.
                tsh.addSchema(sch1);
                assertEquals(sch0table0.getTargetKeys().size(), 1);
                assertEquals(sch1table0.getFromKeys().size(), 1);

                // Case 2C: removing sch1 and re-adding a new instance of sch1
                // In this case the sch1.table0 it isn't in sch1 instance, so
                // there shouldn't be any keys to add.
                tsh.removeSchema("sch1");
                sch1 = tsh.getNewSchema("sch1");
                tsh.addSchema(sch1);
                assertTrue(sch0table0.getTargetKeys().isEmpty());
                assertTrue(sch1table0.getFromKeys().isEmpty());

                //
                // CASE 3: foreign key with multiple columns
                //
                TableEntity tableX = tsh.getNewTable("sch0", "tableX");
                tsh.addTable("sch0", tableX);
                TableEntity tableY = tsh.getNewTable("sch0", "tableY");
                tsh.addTable("sch0", tableY);

                assertTrue(tableX.getFromKeys().isEmpty());
                assertEquals(tableX.getTargetKeys().size(), 1);
                for (KeyColumnEntity keyColumn : tableX.getTargetKeys().get(0).getKeyColumns()) {
                    String fromColumn = keyColumn.getFromColumn();
                    String targetColumn = keyColumn.getFromColumn();
                    assertTrue(fromColumn.equals("idy1") || fromColumn.equals("idy2"));
                    assertTrue(targetColumn.equals("idx1") || targetColumn.equals("idx2"));
                }

                assertTrue(tableY.getTargetKeys().isEmpty());
                assertEquals(tableY.getFromKeys().size(), 1);
                for (KeyColumnEntity keyColumn : tableY.getFromKeys().get(0).getKeyColumns()) {
                    String fromColumn = keyColumn.getFromColumn();
                    String targetColumn = keyColumn.getFromColumn();
                    assertTrue(fromColumn.equals("idy1") || fromColumn.equals("idy2"));
                    assertTrue(targetColumn.equals("idx1") || targetColumn.equals("idx2"));
                }

                /////////////////////////////////////
                //         ADDING A COLUMN         //
                /////////////////////////////////////
                ColumnEntity columnEntity = tsh.getColumnEntities("sch0", "table1").get(0);
                tsh.addColumn(sch0table1, columnEntity);
                assertNotNull(columnEntity.getTable());

                /////////////////////////////////////
                //              SAVE               //
                /////////////////////////////////////
                tsh.save();
                assertNotNull(sch0.getTable("sch0.table0"));

                /////////////////////////////////////
                //           LOAD SAVED            //
                /////////////////////////////////////
                tsh = new TapSchemaHandler(dbWrapper, "test_tap_schema", true);
                sch0 = tsh.getSchemas().iterator().next();
                tsh.removeTable(sch0, sch0.getTable("sch0.table0"));
                assertNull(sch0.getTable("sch0.table0"));
                tsh.save();

                tsh = new TapSchemaHandler(dbWrapper, "test_tap_schema", true);
                sch0 = tsh.getSchemas().iterator().next();
                assertNull(sch0.getTable("sch0.table0"));

                tsh.addTable("sch0", sch0table0);
            }
        } catch (SQLException e) {
            throw e;
        } finally {
            //removeTestingDatabases(credentials);
        }
    }
}
